# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

numba = pytest.importorskip("numba")
ak_connect_numba_arrayview = pytest.importorskip("awkward._connect._numba.arrayview")
ak_connect_numba_layout = pytest.importorskip("awkward._connect._numba.layout")


def test_numpyarray():
    layout = ak.from_iter(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], highlevel=False
    )

    numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
    assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

    lookup1 = ak_connect_numba_arrayview.Lookup(layout)
    lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
    numbatype.form_fill(0, layout, lookup2)

    assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
    assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

    counter = [0]

    def materialize():
        counter[0] += 1
        return layout

    generator = ak.layout.ArrayGenerator(
        materialize, form=layout.form, length=len(layout)
    )
    virtualarray = ak.layout.VirtualArray(generator)

    lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
    assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

    array = ak.Array(virtualarray)
    array.numba_type
    assert counter[0] == 0

    @numba.njit
    def f1(x):
        return x[5]

    assert f1(array) == 5.5
    assert counter[0] == 1

    assert f1(array) == 5.5
    assert counter[0] == 1


def test_listarray():
    for case in "ListOffsetArray", "ListArray":
        layout = ak.from_iter(
            [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], highlevel=False
        )
        if case == "ListArray":
            layout = layout[[0, 1, 2, 3, 4]]

        numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
        assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

        lookup1 = ak_connect_numba_arrayview.Lookup(layout)
        lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
        numbatype.form_fill(0, layout, lookup2)

        assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
        assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

        counter = [0]

        def materialize():
            counter[0] += 1
            return layout

        generator = ak.layout.ArrayGenerator(
            materialize, form=layout.form, length=len(layout)
        )
        virtualarray = ak.layout.VirtualArray(generator)

        lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
        assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

        array = ak.Array(virtualarray)
        array.numba_type
        assert counter[0] == 0

        @numba.njit
        def f3(x):
            return x

        assert isinstance(f3(array).layout, ak.layout.VirtualArray)
        assert counter[0] == 0

        @numba.njit
        def f1(x):
            return x[2][1]

        assert f1(array) == 5.5
        assert counter[0] == 1

        assert f1(array) == 5.5
        assert counter[0] == 1

        @numba.njit
        def f2(x):
            return x[2]

        assert ak.to_list(f2(array)) == [4.4, 5.5]
        assert counter[0] == 1

        assert ak.to_list(f2(array)) == [4.4, 5.5]
        assert counter[0] == 1

        assert ak.to_list(f3(array)) == [
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
            [6.6],
            [7.7, 8.8, 9.9],
        ]


def test_regulararray():
    layout = ak.from_numpy(
        np.array([[1, 2, 3], [4, 5, 6]]), regulararray=True, highlevel=False
    )

    numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
    assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

    lookup1 = ak_connect_numba_arrayview.Lookup(layout)
    lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
    numbatype.form_fill(0, layout, lookup2)

    assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
    assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

    counter = [0]

    def materialize():
        counter[0] += 1
        return layout

    generator = ak.layout.ArrayGenerator(
        materialize, form=layout.form, length=len(layout)
    )
    virtualarray = ak.layout.VirtualArray(generator)

    lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
    assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

    array = ak.Array(virtualarray)
    array.numba_type
    assert counter[0] == 0

    @numba.njit
    def f3(x):
        return x

    assert isinstance(f3(array).layout, ak.layout.VirtualArray)
    assert counter[0] == 0

    @numba.njit
    def f1(x):
        return x[1][1]

    assert f1(array) == 5
    assert counter[0] == 1

    assert f1(array) == 5
    assert counter[0] == 1

    @numba.njit
    def f2(x):
        return x[1]

    assert ak.to_list(f2(array)) == [4, 5, 6]
    assert counter[0] == 1

    assert ak.to_list(f2(array)) == [4, 5, 6]
    assert counter[0] == 1

    assert ak.to_list(f3(array)) == [[1, 2, 3], [4, 5, 6]]


def test_indexedarray():
    layout = ak.from_iter(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], highlevel=False
    )
    layout = ak.layout.IndexedArray64(
        ak.layout.Index64(np.array([4, 3, 2, 1, 0], dtype=np.int64)), layout
    )

    numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
    assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

    lookup1 = ak_connect_numba_arrayview.Lookup(layout)
    lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
    numbatype.form_fill(0, layout, lookup2)

    assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
    assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

    counter = [0]

    def materialize():
        counter[0] += 1
        return layout

    generator = ak.layout.ArrayGenerator(
        materialize, form=layout.form, length=len(layout)
    )
    virtualarray = ak.layout.VirtualArray(generator)

    lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
    assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

    array = ak.Array(virtualarray)
    array.numba_type
    assert counter[0] == 0

    @numba.njit
    def f3(x):
        return x

    assert isinstance(f3(array).layout, ak.layout.VirtualArray)
    assert counter[0] == 0

    @numba.njit
    def f1(x):
        return x[4][1]

    assert f1(array) == 2.2
    assert counter[0] == 1

    assert f1(array) == 2.2
    assert counter[0] == 1

    @numba.njit
    def f2(x):
        return x[4]

    assert ak.to_list(f2(array)) == [1.1, 2.2, 3.3]
    assert counter[0] == 1

    assert ak.to_list(f2(array)) == [1.1, 2.2, 3.3]
    assert counter[0] == 1

    assert ak.to_list(f3(array)) == [
        [7.7, 8.8, 9.9],
        [6.6],
        [4.4, 5.5],
        [],
        [1.1, 2.2, 3.3],
    ]


def test_indexedoptionarray():
    layout = ak.from_iter(
        [[1.1, 2.2, 3.3], None, [], [4.4, 5.5], None, None, [6.6], [7.7, 8.8, 9.9]],
        highlevel=False,
    )

    numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
    assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

    lookup1 = ak_connect_numba_arrayview.Lookup(layout)
    lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
    numbatype.form_fill(0, layout, lookup2)

    assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
    assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

    counter = [0]

    def materialize():
        counter[0] += 1
        return layout

    generator = ak.layout.ArrayGenerator(
        materialize, form=layout.form, length=len(layout)
    )
    virtualarray = ak.layout.VirtualArray(generator)

    lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
    assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

    array = ak.Array(virtualarray)
    array.numba_type
    assert counter[0] == 0

    @numba.njit
    def f3(x):
        return x

    assert isinstance(f3(array).layout, ak.layout.VirtualArray)
    assert counter[0] == 0

    @numba.njit
    def f1(x):
        return x[3][1]

    assert f1(array) == 5.5
    assert counter[0] == 1

    assert f1(array) == 5.5
    assert counter[0] == 1

    @numba.njit
    def f2(x, i):
        return x[i]

    assert ak.to_list(f2(array, 3)) == [4.4, 5.5]
    assert counter[0] == 1

    assert ak.to_list(f2(array, 4)) == None
    assert counter[0] == 1

    assert ak.to_list(f2(array, 3)) == [4.4, 5.5]
    assert counter[0] == 1

    assert ak.to_list(f2(array, 4)) == None
    assert counter[0] == 1

    assert ak.to_list(f3(array)) == [
        [1.1, 2.2, 3.3],
        None,
        [],
        [4.4, 5.5],
        None,
        None,
        [6.6],
        [7.7, 8.8, 9.9],
    ]


def test_bytemaskedarray():
    layout = ak.from_iter(
        [
            [1.1, 2.2, 3.3],
            [999],
            [],
            [4.4, 5.5],
            [123, 321],
            [],
            [6.6],
            [7.7, 8.8, 9.9],
        ],
        highlevel=False,
    )
    layout = ak.layout.ByteMaskedArray(
        ak.layout.Index8(
            np.array(
                [False, True, False, False, True, True, False, False], dtype=np.int8
            )
        ),
        layout,
        valid_when=False,
    )

    numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
    assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

    lookup1 = ak_connect_numba_arrayview.Lookup(layout)
    lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
    numbatype.form_fill(0, layout, lookup2)

    assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
    assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

    counter = [0]

    def materialize():
        counter[0] += 1
        return layout

    generator = ak.layout.ArrayGenerator(
        materialize, form=layout.form, length=len(layout)
    )
    virtualarray = ak.layout.VirtualArray(generator)

    lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
    assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

    array = ak.Array(virtualarray)
    array.numba_type
    assert counter[0] == 0

    @numba.njit
    def f3(x):
        return x

    assert isinstance(f3(array).layout, ak.layout.VirtualArray)
    assert counter[0] == 0

    @numba.njit
    def f1(x):
        return x[3][1]

    assert f1(array) == 5.5
    assert counter[0] == 1

    assert f1(array) == 5.5
    assert counter[0] == 1

    @numba.njit
    def f2(x, i):
        return x[i]

    assert ak.to_list(f2(array, 3)) == [4.4, 5.5]
    assert counter[0] == 1

    assert ak.to_list(f2(array, 4)) == None
    assert counter[0] == 1

    assert ak.to_list(f2(array, 3)) == [4.4, 5.5]
    assert counter[0] == 1

    assert ak.to_list(f2(array, 4)) == None
    assert counter[0] == 1

    assert ak.to_list(f3(array)) == [
        [1.1, 2.2, 3.3],
        None,
        [],
        [4.4, 5.5],
        None,
        None,
        [6.6],
        [7.7, 8.8, 9.9],
    ]


def test_bitmaskedarray():
    layout = ak.from_iter(
        [
            [1.1, 2.2, 3.3],
            [999],
            [],
            [4.4, 5.5],
            [123, 321],
            [],
            [6.6],
            [7.7, 8.8, 9.9],
            [3, 2, 1],
        ],
        highlevel=False,
    )
    layout = ak.layout.BitMaskedArray(
        ak.layout.IndexU8(
            np.packbits(
                np.array(
                    [
                        False,
                        True,
                        False,
                        False,
                        True,
                        True,
                        False,
                        False,
                        False,
                        True,
                        True,
                        True,
                        True,
                        True,
                        True,
                        True,
                    ],
                    dtype=np.bool,
                )
            )
        ),
        layout,
        valid_when=False,
        length=9,
        lsb_order=False,
    )

    numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
    assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

    lookup1 = ak_connect_numba_arrayview.Lookup(layout)
    lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
    numbatype.form_fill(0, layout, lookup2)

    assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
    assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

    counter = [0]

    def materialize():
        counter[0] += 1
        return layout

    generator = ak.layout.ArrayGenerator(
        materialize, form=layout.form, length=len(layout)
    )
    virtualarray = ak.layout.VirtualArray(generator)

    lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
    assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

    array = ak.Array(virtualarray)
    array.numba_type
    assert counter[0] == 0

    @numba.njit
    def f3(x):
        return x

    assert isinstance(f3(array).layout, ak.layout.VirtualArray)
    assert counter[0] == 0

    @numba.njit
    def f1(x):
        return x[3][1]

    assert f1(array) == 5.5
    assert counter[0] == 1

    assert f1(array) == 5.5
    assert counter[0] == 1

    @numba.njit
    def f2(x, i):
        return x[i]

    assert ak.to_list(f2(array, 3)) == [4.4, 5.5]
    assert counter[0] == 1

    assert ak.to_list(f2(array, 4)) == None
    assert counter[0] == 1

    assert ak.to_list(f2(array, 3)) == [4.4, 5.5]
    assert counter[0] == 1

    assert ak.to_list(f2(array, 4)) == None
    assert counter[0] == 1

    assert ak.to_list(f3(array)) == [
        [1.1, 2.2, 3.3],
        None,
        [],
        [4.4, 5.5],
        None,
        None,
        [6.6],
        [7.7, 8.8, 9.9],
        [3.0, 2.0, 1.0],
    ]


def test_unmaskedarray():
    layout = ak.from_iter(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], highlevel=False
    )
    layout = ak.layout.UnmaskedArray(layout)

    numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
    assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

    lookup1 = ak_connect_numba_arrayview.Lookup(layout)
    lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
    numbatype.form_fill(0, layout, lookup2)

    assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
    assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

    counter = [0]

    def materialize():
        counter[0] += 1
        return layout

    generator = ak.layout.ArrayGenerator(
        materialize, form=layout.form, length=len(layout)
    )
    virtualarray = ak.layout.VirtualArray(generator)

    lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
    assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

    array = ak.Array(virtualarray)
    array.numba_type
    assert counter[0] == 0

    @numba.njit
    def f3(x):
        return x

    assert isinstance(f3(array).layout, ak.layout.VirtualArray)
    assert counter[0] == 0

    @numba.njit
    def f1(x):
        return x[2][1]

    assert f1(array) == 5.5
    assert counter[0] == 1

    assert f1(array) == 5.5
    assert counter[0] == 1

    @numba.njit
    def f2(x):
        return x[2]

    assert ak.to_list(f2(array)) == [4.4, 5.5]
    assert counter[0] == 1

    assert ak.to_list(f2(array)) == [4.4, 5.5]
    assert counter[0] == 1

    assert ak.to_list(f3(array)) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]


def test_recordarray():
    layout = ak.from_iter(
        [
            {"x": 0.0, "y": []},
            {"x": 1.1, "y": [1]},
            {"x": 2.2, "y": [2, 2]},
            {"x": 3.3, "y": [3, 3, 3]},
            {"x": 4.4, "y": [4, 4, 4, 4]},
        ],
        highlevel=False,
    )

    numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
    assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

    lookup1 = ak_connect_numba_arrayview.Lookup(layout)
    lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
    numbatype.form_fill(0, layout, lookup2)

    assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
    assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

    counter = [0]

    def materialize():
        counter[0] += 1
        return layout

    generator = ak.layout.ArrayGenerator(
        materialize, form=layout.form, length=len(layout)
    )
    virtualarray = ak.layout.VirtualArray(generator)

    lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
    assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

    array = ak.Array(virtualarray)
    array.numba_type
    assert counter[0] == 0

    @numba.njit
    def f3(x):
        return x

    assert isinstance(f3(array).layout, ak.layout.VirtualArray)
    assert counter[0] == 0

    @numba.njit
    def f1a(x):
        return x["x"][2]

    assert f1a(array) == 2.2
    assert counter[0] == 1

    assert f1a(array) == 2.2
    assert counter[0] == 1

    @numba.njit
    def f1b(x):
        return x[2]["x"]

    assert f1b(array) == 2.2
    assert counter[0] == 1

    assert f1b(array) == 2.2
    assert counter[0] == 1

    @numba.njit
    def f1c(x):
        return x.x[2]

    assert f1c(array) == 2.2
    assert counter[0] == 1

    assert f1c(array) == 2.2
    assert counter[0] == 1

    @numba.njit
    def f1d(x):
        return x[2].x

    assert f1d(array) == 2.2
    assert counter[0] == 1

    assert f1d(array) == 2.2
    assert counter[0] == 1

    @numba.njit
    def f2a(x):
        return x["y"][2]

    assert ak.to_list(f2a(array)) == [2, 2]
    assert counter[0] == 1

    assert ak.to_list(f2a(array)) == [2, 2]
    assert counter[0] == 1

    @numba.njit
    def f2b(x):
        return x[2]["y"]

    assert ak.to_list(f2b(array)) == [2, 2]
    assert counter[0] == 1

    assert ak.to_list(f2b(array)) == [2, 2]
    assert counter[0] == 1

    @numba.njit
    def f2c(x):
        return x.y[2]

    assert ak.to_list(f2a(array)) == [2, 2]
    assert counter[0] == 1

    assert ak.to_list(f2a(array)) == [2, 2]
    assert counter[0] == 1

    @numba.njit
    def f2d(x):
        return x[2].y

    assert ak.to_list(f2b(array)) == [2, 2]
    assert counter[0] == 1

    assert ak.to_list(f2b(array)) == [2, 2]
    assert counter[0] == 1

    assert ak.to_list(f3(array)) == [
        {"x": 0.0, "y": []},
        {"x": 1.1, "y": [1]},
        {"x": 2.2, "y": [2, 2]},
        {"x": 3.3, "y": [3, 3, 3]},
        {"x": 4.4, "y": [4, 4, 4, 4]},
    ]


def test_tuplearray():
    layout = ak.from_iter(
        [(0.0, []), (1.1, [1]), (2.2, [2, 2]), (3.3, [3, 3, 3]), (4.4, [4, 4, 4, 4])],
        highlevel=False,
    )

    numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
    assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

    lookup1 = ak_connect_numba_arrayview.Lookup(layout)
    lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
    numbatype.form_fill(0, layout, lookup2)

    assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
    assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

    counter = [0]

    def materialize():
        counter[0] += 1
        return layout

    generator = ak.layout.ArrayGenerator(
        materialize, form=layout.form, length=len(layout)
    )
    virtualarray = ak.layout.VirtualArray(generator)

    lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
    assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

    array = ak.Array(virtualarray)
    array.numba_type
    assert counter[0] == 0

    @numba.njit
    def f3(x):
        return x

    assert isinstance(f3(array).layout, ak.layout.VirtualArray)
    assert counter[0] == 0

    @numba.njit
    def f1a(x):
        return x["0"][2]

    assert f1a(array) == 2.2
    assert counter[0] == 1

    assert f1a(array) == 2.2
    assert counter[0] == 1

    @numba.njit
    def f1b(x):
        return x[2]["0"]

    assert f1b(array) == 2.2
    assert counter[0] == 1

    assert f1b(array) == 2.2
    assert counter[0] == 1

    @numba.njit
    def f2a(x):
        return x["1"][2]

    assert ak.to_list(f2a(array)) == [2, 2]
    assert counter[0] == 1

    assert ak.to_list(f2a(array)) == [2, 2]
    assert counter[0] == 1

    @numba.njit
    def f2b(x):
        return x[2]["1"]

    assert ak.to_list(f2b(array)) == [2, 2]
    assert counter[0] == 1

    assert ak.to_list(f2b(array)) == [2, 2]
    assert counter[0] == 1

    assert ak.to_list(f3(array)) == [
        (0.0, []),
        (1.1, [1]),
        (2.2, [2, 2]),
        (3.3, [3, 3, 3]),
        (4.4, [4, 4, 4, 4]),
    ]


def test_unionarray():
    layout = ak.from_iter(
        [0.0, [], 1.1, [1], 2.2, [2, 2], 3.3, [3, 3, 3], 4.4, [4, 4, 4, 4]],
        highlevel=False,
    )

    numbatype = ak._connect._numba.arrayview.tonumbatype(layout.form)
    assert ak_connect_numba_layout.typeof(layout).name == numbatype.name

    lookup1 = ak_connect_numba_arrayview.Lookup(layout)
    lookup2 = ak_connect_numba_arrayview.Lookup(layout.form)
    numbatype.form_fill(0, layout, lookup2)

    assert np.array_equal(lookup1.arrayptrs, lookup2.arrayptrs)
    assert np.array_equal(lookup1.sharedptrs == -1, lookup2.sharedptrs == -1)

    counter = [0]

    def materialize():
        counter[0] += 1
        return layout

    generator = ak.layout.ArrayGenerator(
        materialize, form=layout.form, length=len(layout)
    )
    virtualarray = ak.layout.VirtualArray(generator)

    lookup3 = ak_connect_numba_arrayview.Lookup(virtualarray)
    assert len(lookup1.arrayptrs) + 3 == len(lookup3.arrayptrs)

    array = ak.Array(virtualarray)
    array.numba_type
    assert counter[0] == 0

    @numba.njit
    def f3(x):
        return x

    assert isinstance(f3(array).layout, ak.layout.VirtualArray)
    assert counter[0] == 0

    assert ak.to_list(f3(array)) == [
        0.0,
        [],
        1.1,
        [1],
        2.2,
        [2, 2],
        3.3,
        [3, 3, 3],
        4.4,
        [4, 4, 4, 4],
    ]


def test_deep_virtualarrays():
    one = ak.from_iter([0.0, 1.1, 2.2, 3.3, 4.4], highlevel=False)
    two = ak.from_iter([[], [1], [2, 2], [3, 3, 3], [4, 4, 4, 4]], highlevel=False)

    counter = [0, 0]

    def materialize1():
        counter[0] += 1
        return one

    generator1 = ak.layout.ArrayGenerator(materialize1, form=one.form, length=len(one))
    vone = ak.layout.VirtualArray(generator1)

    def materialize2():
        counter[1] += 1
        return two

    generator2 = ak.layout.ArrayGenerator(materialize2, form=two.form, length=len(two))
    vtwo = ak.layout.VirtualArray(generator2)

    recordarray = ak.layout.RecordArray([vone, vtwo], ["x", "y"])
    array = ak.Array(recordarray)
    array.numba_type
    assert counter == [0, 0]

    @numba.njit
    def f3(x):
        return x

    tmp1 = f3(array).layout
    assert isinstance(tmp1, ak.layout.RecordArray)
    assert isinstance(tmp1.field(0), ak.layout.VirtualArray)
    assert isinstance(tmp1.field(1), ak.layout.VirtualArray)
    assert counter == [0, 0]

    @numba.njit
    def f1a(x):
        return x["x"]

    tmp2a = f1a(array).layout
    assert isinstance(tmp2a, ak.layout.VirtualArray)
    assert counter == [0, 0]

    @numba.njit
    def f1b(x):
        return x["x"][2]

    tmp2b = f1b(array)
    assert tmp2b == 2.2
    assert counter == [1, 0]

    @numba.njit
    def f2a(x):
        return x["y"]

    tmp3a = f2a(array).layout
    assert isinstance(tmp3a, ak.layout.VirtualArray)
    assert counter == [1, 0]

    @numba.njit
    def f2b(x):
        return x["y"][2]

    tmp3b = f2b(array).layout
    assert isinstance(tmp3b, ak.layout.NumpyArray)
    assert ak.to_list(tmp3b) == [2, 2]
    assert counter == [1, 1]


def test_nested_virtualness():
    counter = [0, 0]

    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )

    def materialize1():
        counter[1] += 1
        return content

    generator1 = ak.layout.ArrayGenerator(
        materialize1, form=content.form, length=len(content)
    )
    virtual1 = ak.layout.VirtualArray(generator1)

    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10], dtype=np.int64))
    listarray = ak.layout.ListOffsetArray64(offsets, virtual1)

    def materialize2():
        counter[0] += 1
        return listarray

    generator2 = ak.layout.ArrayGenerator(
        materialize2, form=listarray.form, length=len(listarray)
    )
    virtual2 = ak.layout.VirtualArray(generator2)
    array = ak.Array(virtual2)

    assert counter == [0, 0]

    @numba.njit
    def f1(x, i):
        return x[i]

    tmp1 = f1(array, 2)
    assert counter == [1, 0]

    tmp2 = f1(tmp1, 1)
    assert tmp2 == 4.4
    assert counter == [1, 1]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_numpyarray():
    array = ak.layout.NumpyArray(np.arange(10) * 1.1)

    with pytest.raises(ValueError) as excinfo:
        array[20]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[-20]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    array[-20:20]

    with pytest.raises(ValueError) as excinfo:
        array[
            20,
        ]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[
            -20,
        ]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    array[
        -20:20,
    ]

    with pytest.raises(ValueError) as excinfo:
        array[2, 3]
    assert ", too many dimensions in slice" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[[5, 3, 20, 8]]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[[5, 3, -20, 8]]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    array.setidentities()

    with pytest.raises(ValueError) as excinfo:
        array[20]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[-20]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    array[-20:20]

    with pytest.raises(ValueError) as excinfo:
        array[
            20,
        ]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[
            -20,
        ]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    array[
        -20:20,
    ]

    with pytest.raises(ValueError) as excinfo:
        array[2, 3]
    assert ", too many dimensions in slice" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[[5, 3, 20, 8]]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[[5, 3, -20, 8]]
    assert " attempting to get -20, index out of range" in str(excinfo.value)


def test_listarray_numpyarray():
    starts = ak.layout.Index64(np.array([0, 3, 3, 5, 6]))
    stops = ak.layout.Index64(np.array([3, 3, 5, 6]))
    content = ak.layout.NumpyArray(np.arange(10) * 1.1)
    with pytest.raises(ValueError) as excinfo:
        array = ak.layout.ListArray64(starts, stops, content)

    starts = ak.layout.Index64(np.array([0, 3, 3, 5, 6]))
    stops = ak.layout.Index64(np.array([3, 3, 5, 6, 10]))
    content = ak.layout.NumpyArray(np.arange(10) * 1.1)
    array = ak.layout.ListArray64(starts, stops, content)

    with pytest.raises(ValueError) as excinfo:
        array[20]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[-20]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    array[-20:20]

    with pytest.raises(ValueError) as excinfo:
        array[
            20,
        ]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[
            -20,
        ]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    array[
        -20:20,
    ]

    with pytest.raises(ValueError) as excinfo:
        array[2, 1, 0]
    assert ", too many dimensions in slice" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[[2, 0, 0, 20, 3]]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array[[2, 0, 0, -20, 3]]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    starts = ak.layout.Index64(np.array([0, 3, 3, 5, 6]))
    stops = ak.layout.Index64(np.array([3, 3, 5, 6, 10]))
    content = ak.layout.NumpyArray(np.arange(10) * 1.1)
    array = ak.layout.ListArray64(starts, stops, content)

    array.setidentities()

    with pytest.raises(ValueError) as excinfo:
        array[2, 20]
    assert " with identity [2] attempting to get 20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array[2, -20]
    assert " with identity [2] attempting to get -20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array[1:][2, 20]
    assert " with identity [3] attempting to get 20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array[1:][2, -20]
    assert " with identity [3] attempting to get -20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array[2, [1, 0, 0, 20]]
    assert " with identity [2] attempting to get 20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array[2, [1, 0, 0, -20]]
    assert " with identity [2] attempting to get -20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array[1:][2, [0, 20]]
    assert " with identity [3] attempting to get 20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array[1:][2, [0, -20]]
    assert " with identity [3] attempting to get -20, index out of range" in str(
        excinfo.value
    )


def test_listarray_listarray_numpyarray():
    content = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts1 = ak.layout.Index64(np.array([0, 3, 3, 5, 6]))
    stops1 = ak.layout.Index64(np.array([3, 3, 5, 6, 9]))
    offsets1 = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9]))
    starts2 = ak.layout.Index64(np.array([0, 2, 3, 3]))
    stops2 = ak.layout.Index64(np.array([2, 3, 3, 5]))
    offsets2 = ak.layout.Index64(np.array([0, 2, 3, 3, 5]))

    array1 = ak.layout.ListArray64(starts1, stops1, content)
    array2 = ak.layout.ListArray64(starts2, stops2, array1)

    with pytest.raises(ValueError) as excinfo:
        array2[20]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array2[
            20,
        ]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array2[2, 20]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array2[-20]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array2[
            -20,
        ]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array2[2, -20]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array2[1, 0, 20]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    array2.setidentities()

    with pytest.raises(ValueError) as excinfo:
        array2[20]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array2[
            20,
        ]
    assert " attempting to get 20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array2[2, 20]
    assert " with identity [2] attempting to get 20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array2[1:][2, 20]
    assert " with identity [3] attempting to get 20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array2[-20]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array2[
            -20,
        ]
    assert " attempting to get -20, index out of range" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        array2[2, -20]
    assert " with identity [2] attempting to get -20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array2[1:][2, -20]
    assert " with identity [3] attempting to get -20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array2[1, 0, 20]
    assert " with identity [1, 0] attempting to get 20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array2[1:][2, 0, 20]
    assert " with identity [3, 0] attempting to get 20, index out of range" in str(
        excinfo.value
    )

    with pytest.raises(ValueError) as excinfo:
        array2[:, 1:][3, 0, 20]
    assert " with identity [3, 1] attempting to get 20, index out of range" in str(
        excinfo.value
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import ctypes
import platform
import pkg_resources

if platform.system() == "Windows":
    name = "awkward-cpu-kernels.dll"
elif platform.system() == "Darwin":
    name = "libawkward-cpu-kernels.dylib"
else:
    name = "libawkward-cpu-kernels.so"
libpath = pkg_resources.resource_filename("awkward", name)

lib = ctypes.cdll.LoadLibrary(libpath)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import json

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    content = ak.Array([1, 2, 3, 4, 5])
    generator1 = ak.layout.ArrayGenerator(
        lambda a, b: a * content + b, (100,), {"b": 3}
    )
    assert generator1.form is None
    assert ak.to_list(generator1()) == [103, 203, 303, 403, 503]
    assert generator1.args == (100,)
    assert generator1.kwargs == {"b": 3}
    assert generator1.form is not None
    assert generator1.length is None

    generator2 = generator1.with_args((1000,))
    assert generator2.form is None
    assert ak.to_list(generator2()) == [1003, 2003, 3003, 4003, 5003]
    assert generator2.args == (1000,)
    assert generator2.kwargs == {"b": 3}
    assert generator2.form is not None
    assert generator2.length is None

    generator3 = generator1.with_kwargs({"b": 4})
    assert generator3.form is None
    assert ak.to_list(generator3()) == [104, 204, 304, 404, 504]
    assert generator3.args == (100,)
    assert generator3.kwargs == {"b": 4}
    assert generator3.form is not None
    assert generator3.length is None

    generator4 = generator1.with_form(ak.forms.Form.fromjson('"int64"'))
    assert generator4.form is not None
    assert ak.to_list(generator4()) == [103, 203, 303, 403, 503]
    assert generator4.args == (100,)
    assert generator4.kwargs == {"b": 3}
    assert json.loads(generator4.form.tojson()) == json.loads(
        ak.forms.Form.fromjson('"int64"').tojson()
    )
    assert generator4.length is None

    generator5 = generator1.with_length(5)
    assert generator5.form is None
    assert ak.to_list(generator5()) == [103, 203, 303, 403, 503]
    assert generator5.args == (100,)
    assert generator5.kwargs == {"b": 3}
    assert generator5.form is not None
    assert generator5.length == 5



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

from awkward._ext import Index8
from awkward._ext import IndexU8
from awkward._ext import Index32
from awkward._ext import IndexU32
from awkward._ext import Index64

from awkward._ext import Identities32
from awkward._ext import Identities64

from awkward._ext import Iterator
from awkward._ext import ArrayBuilder
from awkward._ext import _PersistentSharedPtr

from awkward._ext import Content

from awkward._ext import EmptyArray

from awkward._ext import IndexedArray32
from awkward._ext import IndexedArrayU32
from awkward._ext import IndexedArray64
from awkward._ext import IndexedOptionArray32
from awkward._ext import IndexedOptionArray64

from awkward._ext import ByteMaskedArray
from awkward._ext import BitMaskedArray
from awkward._ext import UnmaskedArray

from awkward._ext import ListArray32
from awkward._ext import ListArrayU32
from awkward._ext import ListArray64

from awkward._ext import ListOffsetArray32
from awkward._ext import ListOffsetArrayU32
from awkward._ext import ListOffsetArray64

from awkward._ext import NumpyArray

from awkward._ext import Record
from awkward._ext import RecordArray

from awkward._ext import RegularArray

from awkward._ext import UnionArray8_32
from awkward._ext import UnionArray8_U32
from awkward._ext import UnionArray8_64

from awkward._ext import VirtualArray
from awkward._ext import ArrayGenerator
from awkward._ext import SliceGenerator
from awkward._ext import ArrayCache

from awkward._ext import kernel_lib



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    assert np.asarray(
        ak.layout.ByteMaskedArray(
            ak.layout.Index8(np.array([1, 1, 1, 1, 1], np.int8)),
            ak.layout.NumpyArray(np.array([0, 1, 2, 3, 4])),
            valid_when=True,
        ).bytemask()
    ).tolist() == [0, 0, 0, 0, 0]

    assert np.asarray(
        ak.layout.ByteMaskedArray(
            ak.layout.Index8(np.array([1, 1, 1, 1, 1], np.int8)),
            ak.layout.NumpyArray(np.array([0, 1, 2, 3, 4])),
            valid_when=False,
        ).bytemask()
    ).tolist() == [1, 1, 1, 1, 1]

    assert np.asarray(
        ak.layout.BitMaskedArray(
            ak.layout.IndexU8(np.array([31], np.uint8)),
            ak.layout.NumpyArray(np.array([0, 1, 2, 3, 4])),
            valid_when=True,
            length=5,
            lsb_order=True,
        ).bytemask()
    ).tolist() == [0, 0, 0, 0, 0]

    assert np.asarray(
        ak.layout.BitMaskedArray(
            ak.layout.IndexU8(np.array([31], np.uint8)),
            ak.layout.NumpyArray(np.array([0, 1, 2, 3, 4])),
            valid_when=False,
            length=5,
            lsb_order=True,
        ).bytemask()
    ).tolist() == [1, 1, 1, 1, 1]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import awkward as ak


def mixin_class(registry):
    """
    Args:
        registry (dict): The destination behavior mapping registry. Typically,
            this would be the global registry #ak.behavior, but one may wish
            to register methods in an alternative way.

    This decorator can be used to register a behavior mixin class.

    Any inherited behaviors will automatically be made available to the decorated
    class.

    See the "Mixin decorators" section of #ak.behavior for further details.
    """

    def register(cls):
        name = cls.__name__
        record = type(
            name + "Record", (cls, ak.highlevel.Record), {"__module__": cls.__module__}
        )
        setattr(sys.modules[cls.__module__], name + "Record", record)
        registry[name] = record
        array = type(
            name + "Array", (cls, ak.highlevel.Array), {"__module__": cls.__module__}
        )
        setattr(sys.modules[cls.__module__], name + "Array", array)
        registry["*", name] = array
        for basecls in cls.mro():
            for method in basecls.__dict__.values():
                if hasattr(method, "_awkward_mixin"):
                    ufunc, rhs, transpose = method._awkward_mixin
                    if rhs is None:
                        registry.setdefault((ufunc, name), method)
                        continue
                    for rhs_name in list(rhs) + [name]:
                        registry.setdefault((ufunc, name, rhs_name), method)
                        if transpose is not None and rhs_name != name:
                            registry.setdefault((ufunc, rhs_name, name), transpose)
                    if basecls.__name__ in rhs:
                        rhs.add(name)
        return cls

    return register


def mixin_class_method(ufunc, rhs=None, transpose=True):
    """
    Args:
        ufunc (numpy.ufunc): A universal function (or NEP18 callable) that is
            hooked in Awkward Array, i.e. it can be the first argument of a behavior.
        rhs (Set[type] or None): Set of right-hand side argument types, optional
            if wrapping a unary function. The left-hand side is expected to
            always be `self` of the parent class.
        transpose (bool): If true, autmatically create a transpose signature
            (only makes sense for binary ufuncs).

    This decorator can be used to register a mixin class method.

    Using this decorator ensures that derived classes that are declared with the
    #ak.mixin_class decorator will also have the behaviors that this class has.
    """

    def register(method):
        if not isinstance(rhs, (set, type(None))):
            raise ValueError(
                "expected a set of right-hand-side argument types"
                + ak._util.exception_suffix(__file__)
            )
        if transpose and rhs is not None:

            def transposed(left, right):
                return method(right, left)

            # make a copy of rhs, we will edit it later
            method._awkward_mixin = (ufunc, set(rhs), transposed)
        else:
            method._awkward_mixin = (ufunc, rhs, None)
        return method

    return register


__all__ = [
    x
    for x in list(globals())
    if not x.startswith("_") and x not in ("absolute_import", "awkward", "sys")
]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


class Point(ak.Record):
    def __repr__(self):
        return "<{} {}>".format(self["x"], self["y"])


def test():
    behavior = {}
    behavior["__typestr__", "Point"] = "P"
    behavior["Point"] = Point
    array = ak.Array(
        [
            [{"x": 1, "y": [1.1]}, {"x": 2, "y": [2.0, 0.2]}],
            [],
            [{"x": 3, "y": [3.0, 0.3, 3.3]}],
        ],
        with_name="Point",
        behavior=behavior,
        check_valid=True,
    )
    assert repr(array[0, 0]) == "<1 [1.1]>"
    assert repr(array[0]) == "<Array [<1 [1.1]>, <2 [2, 0.2]>] type='2 * P'>"
    assert (
        repr(array) == "<Array [[<1 [1.1]>, ... <3 [3, 0.3, 3.3]>]] type='3 * var * P'>"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    out = ak.ArrayBuilder()

    out.begin_record()
    if True:
        out.field("x")
        out.integer(3)
        out.field("extreme")
        out.begin_record()
        if True:
            out.field("pt")
            out.real(3.3)
            out.field("charge")
            out.integer(-1)
            out.field("iso")
            out.integer(100)
        out.end_record()
    out.end_record()

    out.begin_record()
    if True:
        out.field("x")
        out.integer(3)
    out.end_record()

    ss = out.snapshot()
    assert ak.to_list(ss) == [
        {"x": 3, "extreme": {"pt": 3.3, "charge": -1, "iso": 100}},
        {"x": 3, "extreme": None},
    ]

    assert ak.to_list(
        ak.Array(
            [{"x": 3, "extreme": {"pt": 3.3, "charge": -1, "iso": 100}}, {"x": 3}],
            check_valid=True,
        )
    ) == [
        {"x": 3, "extreme": {"pt": 3.3, "charge": -1, "iso": 100}},
        {"x": 3, "extreme": None},
    ]

    assert ak.to_list(
        ak.Array(
            [
                {"x": 3, "extreme": {"pt": 3.3, "charge": -1, "iso": 100}},
                {"x": 3, "what": 3},
            ],
            check_valid=True,
        )
    ) == [
        {"x": 3, "extreme": {"pt": 3.3, "charge": -1, "iso": 100}, "what": None},
        {"x": 3, "extreme": None, "what": 3},
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


class Canary(dict):
    def __init__(self):
        super(Canary, self).__init__()
        self.ops = []

    def __getitem__(self, key):
        self.ops.append(("get", key))
        return super(Canary, self).__getitem__(key)

    def __setitem__(self, key, value):
        self.ops.append(("set", key))
        return super(Canary, self).__setitem__(key, value)


def test_lazy_arrayset():
    array = ak.from_json(
        """
    [
        {
            "listcollection": [
                {"item1": 1, "item2": 2},
                {"item1": 2, "item2": 4},
                {"item1": 3, "item2": 6}
            ],
            "collection": {"item1": 3, "item2": 4},
            "singleton": 5,
            "listsingleton": [1, 2, 3],
            "unioncollection": {"item1": 3},
            "masked": null
        },
        {
            "listcollection": [
                {"item1": 1, "item2": 2},
                {"item1": 2, "item2": 4},
                {"item1": 3, "item2": 6}
            ],
            "collection": {"item1": 3, "item2": 4},
            "singleton": 5,
            "listsingleton": [1, 2, 3],
            "unioncollection": [{"item1": 2}],
            "masked": 4
        },
        {
            "listcollection": [
                {"item1": 1, "item2": 2},
                {"item1": 2, "item2": 4},
                {"item1": 3, "item2": 6}
            ],
            "collection": {"item1": 3, "item2": 4},
            "singleton": 5,
            "listsingleton": [1, 2, 3],
            "unioncollection": {"item1": 4},
            "masked": 4
        }
    ]"""
    )

    canary = Canary()
    prefix = "kitty"
    form, container, npart = ak.to_arrayset(array, container=canary, prefix=prefix)
    assert not any(op[0] == "get" for op in canary.ops)
    canary.ops = []

    cache = {}
    out = ak.from_arrayset(
        form,
        container,
        lazy=True,
        lazy_cache=cache,
        lazy_lengths=3,
        prefix=prefix,
        lazy_cache_key="hello",
    )
    assert len(canary.ops) == 0
    assert len(cache) == 0

    assert len(out) == 3
    assert len(canary.ops) == 0
    assert len(cache) == 0

    assert ak.to_list(ak.num(out.listcollection)) == [3, 3, 3]
    assert set(canary.ops) == {("get", "kitty-node1-offsets")}
    assert set(cache) == {"hello", "hello-kitty-node1-virtual"}
    canary.ops = []
    cache.clear()

    assert ak.to_list(out.unioncollection) == [
        {"item1": 3},
        [{"item1": 2}],
        {"item1": 4},
    ]
    assert set(canary.ops) == {
        ("get", "kitty-node11-tags"),
        ("get", "kitty-node11-index"),
        ("get", "kitty-node14-offsets"),
        ("get", "kitty-node13"),
        ("get", "kitty-node16"),
    }
    assert set(cache) == {
        "hello",
        "hello-kitty-node11-virtual",
        "hello-kitty-node13-virtual",
        "hello-kitty-node16-virtual",
    }
    canary.ops = []
    cache.clear()

    assert ak.to_list(out.masked) == [None, 4, 4]
    assert set(canary.ops) == {("get", "kitty-node17-index"), ("get", "kitty-node18")}
    assert set(cache) == {"hello", "hello-kitty-node17-virtual"}
    canary.ops = []
    cache.clear()


def test_longer_than_expected():
    array = ak.Array(
        ak.layout.ListOffsetArray64(
            ak.layout.Index64([0, 2, 4]),
            ak.layout.RecordArray(
                {
                    "item1": ak.layout.NumpyArray(np.arange(4)),
                    "longitem": ak.layout.NumpyArray(np.arange(6)),
                }
            ),
        )
    )
    out = ak.from_arrayset(*ak.to_arrayset(array), lazy=True, lazy_lengths=2)
    assert ak.to_list(out) == [
        [{"item1": 0, "longitem": 0}, {"item1": 1, "longitem": 1}],
        [{"item1": 2, "longitem": 2}, {"item1": 3, "longitem": 3}],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array([[1, 2, 3], [], [4, 5], [6, 7, 8, 9]])
    assert ak.flatten(array[:-1], axis=1).tolist() == [1, 2, 3, 4, 5]
    assert ak.flatten(array[:-2], axis=1).tolist() == [1, 2, 3]
    assert ak.flatten(array[:-1], axis=None).tolist() == [1, 2, 3, 4, 5]
    assert ak.flatten(array[:-2], axis=None).tolist() == [1, 2, 3]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    for itype in ["i8", "u8", "i32", "u32", "i64"]:
        form = ak.forms.ListOffsetForm(itype, ak.forms.EmptyForm())
        assert form.offsets == itype



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_typeempty():
    assert np.asarray(ak.types.PrimitiveType("bool").empty()).dtype == np.dtype("bool")
    assert np.asarray(ak.types.PrimitiveType("int8").empty()).dtype == np.dtype("int8")
    assert np.asarray(ak.types.PrimitiveType("int16").empty()).dtype == np.dtype(
        "int16"
    )
    assert np.asarray(ak.types.PrimitiveType("int32").empty()).dtype == np.dtype(
        "int32"
    )
    assert np.asarray(ak.types.PrimitiveType("int64").empty()).dtype == np.dtype(
        "int64"
    )
    assert np.asarray(ak.types.PrimitiveType("uint8").empty()).dtype == np.dtype(
        "uint8"
    )
    assert np.asarray(ak.types.PrimitiveType("uint16").empty()).dtype == np.dtype(
        "uint16"
    )
    assert np.asarray(ak.types.PrimitiveType("uint32").empty()).dtype == np.dtype(
        "uint32"
    )
    assert np.asarray(ak.types.PrimitiveType("uint64").empty()).dtype == np.dtype(
        "uint64"
    )
    assert np.asarray(ak.types.PrimitiveType("float32").empty()).dtype == np.dtype(
        "float32"
    )
    assert np.asarray(ak.types.PrimitiveType("float64").empty()).dtype == np.dtype(
        "float64"
    )
    assert type(ak.types.UnknownType().empty()) is ak.layout.EmptyArray
    assert (
        type(ak.types.ArrayType(ak.types.UnknownType(), 0).empty())
        is ak.layout.EmptyArray
    )
    assert (
        type(ak.types.RegularType(ak.types.UnknownType(), 5).empty())
        is ak.layout.RegularArray
    )
    assert (
        type(ak.types.ListType(ak.types.UnknownType()).empty())
        is ak.layout.ListOffsetArray64
    )
    array = ak.types.RecordType(
        {
            "one": ak.types.PrimitiveType("float64"),
            "two": ak.types.ListType(ak.types.PrimitiveType("float64")),
        }
    ).empty()
    assert type(array) is ak.layout.RecordArray
    assert type(array["one"]) is ak.layout.NumpyArray
    assert np.asarray(array["one"]).dtype == np.dtype("float64")
    assert type(array["two"]) is ak.layout.ListOffsetArray64



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_identity():
    a = np.arange(10)
    b = ak.layout.NumpyArray(a)
    b.setidentities()
    assert np.array(b.identities).tolist() == np.arange(10).reshape(-1, 1).tolist()

    assert np.array(b[3]) == a[3]
    assert (
        np.array(b[3:7].identities).tolist()
        == np.arange(10).reshape(-1, 1)[3:7].tolist()
    )
    assert (
        np.array(b[[7, 3, 3, -4]].identities).tolist()
        == np.arange(10).reshape(-1, 1)[[7, 3, 3, -4]].tolist()
    )
    assert (
        np.array(
            b[
                [True, True, True, False, False, False, True, False, True, False]
            ].identities
        ).tolist()
        == np.arange(10)
        .reshape(-1, 1)[
            [True, True, True, False, False, False, True, False, True, False]
        ]
        .tolist()
    )

    assert np.array(b[1:][3]) == a[1:][3]
    assert (
        np.array(b[1:][3:7].identities).tolist()
        == np.arange(10).reshape(-1, 1)[1:][3:7].tolist()
    )
    assert (
        np.array(b[1:][[7, 3, 3, -4]].identities).tolist()
        == np.arange(10).reshape(-1, 1)[1:][[7, 3, 3, -4]].tolist()
    )
    assert (
        np.array(
            b[1:][
                [True, True, False, False, False, True, False, True, False]
            ].identities
        ).tolist()
        == np.arange(10)
        .reshape(-1, 1)[1:][[True, True, False, False, False, True, False, True, False]]
        .tolist()
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_basic():
    one = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.from_iter([[6.6], [], [], [], [7.7, 8.8, 9.9]], highlevel=False)
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    assert len(array) == 8
    assert [ak.to_list(x) for x in array.partitions] == [
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]],
        [[6.6], [], [], [], [7.7, 8.8, 9.9]],
    ]
    assert ak.to_list(array.partition(0)) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert ak.to_list(array.partition(1)) == [[6.6], [], [], [], [7.7, 8.8, 9.9]]
    assert array.start(0) == 0
    assert array.start(1) == 3
    assert array.stop(0) == 3
    assert array.stop(1) == 8
    assert array.partitionid_index_at(0) == (0, 0)
    assert array.partitionid_index_at(1) == (0, 1)
    assert array.partitionid_index_at(2) == (0, 2)
    assert array.partitionid_index_at(3) == (1, 0)
    assert array.partitionid_index_at(4) == (1, 1)
    assert array.partitionid_index_at(5) == (1, 2)
    assert array.partitionid_index_at(6) == (1, 3)
    assert array.partitionid_index_at(7) == (1, 4)

    assert array.tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"

    assert ak.to_list(array._ext.getitem_at(0)) == [1.1, 2.2, 3.3]
    assert ak.to_list(array._ext.getitem_at(1)) == []
    assert ak.to_list(array._ext.getitem_at(2)) == [4.4, 5.5]
    assert ak.to_list(array._ext.getitem_at(3)) == [6.6]
    assert ak.to_list(array._ext.getitem_at(4)) == []
    assert ak.to_list(array._ext.getitem_at(5)) == []
    assert ak.to_list(array._ext.getitem_at(6)) == []
    assert ak.to_list(array._ext.getitem_at(7)) == [7.7, 8.8, 9.9]
    assert ak.to_list(array._ext.getitem_at(-1)) == [7.7, 8.8, 9.9]
    assert ak.to_list(array._ext.getitem_at(-2)) == []
    assert ak.to_list(array._ext.getitem_at(-3)) == []
    assert ak.to_list(array._ext.getitem_at(-4)) == []
    assert ak.to_list(array._ext.getitem_at(-5)) == [6.6]
    assert ak.to_list(array._ext.getitem_at(-6)) == [4.4, 5.5]
    assert ak.to_list(array._ext.getitem_at(-7)) == []
    assert ak.to_list(array._ext.getitem_at(-8)) == [1.1, 2.2, 3.3]

    assert (
        array._ext.getitem_range(0, 8, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    )
    assert (
        array._ext.getitem_range(1, 8, 1).tojson()
        == "[[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    )
    assert (
        array._ext.getitem_range(2, 8, 1).tojson()
        == "[[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    )
    assert (
        array._ext.getitem_range(3, 8, 1).tojson() == "[[6.6],[],[],[],[7.7,8.8,9.9]]"
    )
    assert array._ext.getitem_range(4, 8, 1).tojson() == "[[],[],[],[7.7,8.8,9.9]]"
    assert array._ext.getitem_range(5, 8, 1).tojson() == "[[],[],[7.7,8.8,9.9]]"
    assert array._ext.getitem_range(6, 8, 1).tojson() == "[[],[7.7,8.8,9.9]]"
    assert array._ext.getitem_range(7, 8, 1).tojson() == "[[7.7,8.8,9.9]]"
    assert array._ext.getitem_range(8, 8, 1).tojson() == "[]"
    assert array._ext.getitem_range(-1, 8, 1).tojson() == "[[7.7,8.8,9.9]]"
    assert array._ext.getitem_range(-2, 8, 1).tojson() == "[[],[7.7,8.8,9.9]]"
    assert array._ext.getitem_range(-3, 8, 1).tojson() == "[[],[],[7.7,8.8,9.9]]"
    assert array._ext.getitem_range(-4, 8, 1).tojson() == "[[],[],[],[7.7,8.8,9.9]]"
    assert (
        array._ext.getitem_range(-5, 8, 1).tojson() == "[[6.6],[],[],[],[7.7,8.8,9.9]]"
    )
    assert (
        array._ext.getitem_range(-6, 8, 1).tojson()
        == "[[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    )
    assert (
        array._ext.getitem_range(-7, 8, 1).tojson()
        == "[[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    )
    assert (
        array._ext.getitem_range(-8, 8, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    )

    assert (
        array._ext.getitem_range(0, 8, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    )
    assert (
        array._ext.getitem_range(0, 7, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[]]"
    )
    assert (
        array._ext.getitem_range(0, 6, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[]]"
    )
    assert (
        array._ext.getitem_range(0, 5, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[]]"
    )
    assert (
        array._ext.getitem_range(0, 4, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6]]"
    )
    assert array._ext.getitem_range(0, 3, 1).tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5]]"
    assert array._ext.getitem_range(0, 2, 1).tojson() == "[[1.1,2.2,3.3],[]]"
    assert array._ext.getitem_range(0, 1, 1).tojson() == "[[1.1,2.2,3.3]]"
    assert array._ext.getitem_range(0, 0, 1).tojson() == "[]"
    assert array._ext.getitem_range(0, -8, 1).tojson() == "[]"
    assert array._ext.getitem_range(0, -7, 1).tojson() == "[[1.1,2.2,3.3]]"
    assert array._ext.getitem_range(0, -6, 1).tojson() == "[[1.1,2.2,3.3],[]]"
    assert array._ext.getitem_range(0, -5, 1).tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5]]"
    assert (
        array._ext.getitem_range(0, -4, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6]]"
    )
    assert (
        array._ext.getitem_range(0, -3, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[]]"
    )
    assert (
        array._ext.getitem_range(0, -2, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[]]"
    )
    assert (
        array._ext.getitem_range(0, -1, 1).tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[]]"
    )

    assert (
        array._ext.getitem_range(0, 8, 2).tojson() == "[[1.1,2.2,3.3],[4.4,5.5],[],[]]"
    )
    assert array._ext.getitem_range(1, 8, 2).tojson() == "[[],[6.6],[],[7.7,8.8,9.9]]"
    assert array._ext.getitem_range(1, 7, 2).tojson() == "[[],[6.6],[]]"
    assert array._ext.getitem_range(2, 8, 2).tojson() == "[[4.4,5.5],[],[]]"
    assert array._ext.getitem_range(3, 8, 2).tojson() == "[[6.6],[],[7.7,8.8,9.9]]"
    assert array._ext.getitem_range(0, 8, 3).tojson() == "[[1.1,2.2,3.3],[6.6],[]]"
    assert array._ext.getitem_range(1, 8, 3).tojson() == "[[],[],[7.7,8.8,9.9]]"
    assert array._ext.getitem_range(1, 7, 3).tojson() == "[[],[]]"
    assert array._ext.getitem_range(2, 8, 3).tojson() == "[[4.4,5.5],[]]"
    assert array._ext.getitem_range(3, 8, 3).tojson() == "[[6.6],[]]"

    assert array._ext.getitem_range(2, 0, -1).tojson() == "[[4.4,5.5],[]]"
    assert (
        array._ext.getitem_range(2, None, -1).tojson() == "[[4.4,5.5],[],[1.1,2.2,3.3]]"
    )
    assert array._ext.getitem_range(1, None, -1).tojson() == "[[],[1.1,2.2,3.3]]"
    assert (
        array._ext.getitem_range(3, None, -1).tojson()
        == "[[6.6],[4.4,5.5],[],[1.1,2.2,3.3]]"
    )
    assert (
        array._ext.getitem_range(None, None, -1).tojson()
        == "[[7.7,8.8,9.9],[],[],[],[6.6],[4.4,5.5],[],[1.1,2.2,3.3]]"
    )
    assert (
        array._ext.getitem_range(-1, None, -1).tojson()
        == "[[7.7,8.8,9.9],[],[],[],[6.6],[4.4,5.5],[],[1.1,2.2,3.3]]"
    )
    assert (
        array._ext.getitem_range(-2, None, -1).tojson()
        == "[[],[],[],[6.6],[4.4,5.5],[],[1.1,2.2,3.3]]"
    )
    assert (
        array._ext.getitem_range(-2, 0, -1).tojson() == "[[],[],[],[6.6],[4.4,5.5],[]]"
    )
    assert array._ext.getitem_range(-2, 1, -1).tojson() == "[[],[],[],[6.6],[4.4,5.5]]"
    assert array._ext.getitem_range(-2, 2, -1).tojson() == "[[],[],[],[6.6]]"
    assert array._ext.getitem_range(-1, 3, -1).tojson() == "[[7.7,8.8,9.9],[],[],[]]"
    assert (
        array._ext.getitem_range(-1, None, -2).tojson() == "[[7.7,8.8,9.9],[],[6.6],[]]"
    )
    assert (
        array._ext.getitem_range(-2, None, -2).tojson()
        == "[[],[],[4.4,5.5],[1.1,2.2,3.3]]"
    )
    assert array._ext.getitem_range(-1, None, -3).tojson() == "[[7.7,8.8,9.9],[],[]]"
    assert array._ext.getitem_range(-2, None, -3).tojson() == "[[],[6.6],[1.1,2.2,3.3]]"
    assert array._ext.getitem_range(-3, None, -3).tojson() == "[[],[4.4,5.5]]"
    assert array._ext.getitem_range(-4, None, -3).tojson() == "[[],[]]"
    assert array._ext.getitem_range(-5, None, -3).tojson() == "[[6.6],[1.1,2.2,3.3]]"
    assert array._ext.getitem_range(-2, 0, -2).tojson() == "[[],[],[4.4,5.5]]"

    assert [ak.to_list(x) for x in array] == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [],
        [],
        [],
        [7.7, 8.8, 9.9],
    ]

    assert ak.to_list(array.toContent()) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [],
        [],
        [],
        [7.7, 8.8, 9.9],
    ]


def test_range_slices():
    a1 = ak.from_iter(np.array([0, 1, 2], dtype=np.int64), highlevel=False)
    a2 = ak.from_iter(np.array([3, 4], dtype=np.int64), highlevel=False)
    a3 = ak.from_iter(np.array([5], dtype=np.int64), highlevel=False)
    a4 = ak.from_iter(np.array([], dtype=np.int64), highlevel=False)
    a5 = ak.from_iter(np.array([6, 7, 8, 9], dtype=np.int64), highlevel=False)
    aspart = ak.partition.IrregularlyPartitionedArray([a1, a2, a3, a4, a5])
    asfull = ak.concatenate([a1, a2, a3, a4, a5], highlevel=False)
    aslist = ak.to_list(asfull)

    for start in range(10):
        for stop in range(10):
            for step in (1, 2, 3, 4, 5, -1, -2, -3, -4, -5):
                assert ak.to_list(asfull[start:stop:step]) == aslist[start:stop:step]
                assert (
                    aspart._ext.getitem_range(start, stop, step).tojson()
                    == asfull[start:stop:step].tojson()
                )


def test_as_slice():
    one = ak.from_iter([False, True, False], highlevel=False)
    two = ak.from_iter([True, True, True, False, False], highlevel=False)
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    target = ak.Array([0, 1, 2, 100, 200, 300, 400, 500])
    assert ak.to_list(target[array]) == [1, 100, 200, 300]


def test_repartition():
    one = ak.from_iter([0, 1, 2], highlevel=False)
    two = ak.from_iter([100, 200, 300, 400, 500], highlevel=False)
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    assert [list(x) for x in array.repartition([3, 8]).partitions] == [
        [0, 1, 2],
        [100, 200, 300, 400, 500],
    ]
    assert [list(x) for x in array.repartition([8]).partitions] == [
        [0, 1, 2, 100, 200, 300, 400, 500]
    ]
    assert [list(x) for x in array.repartition([4, 5, 8]).partitions] == [
        [0, 1, 2, 100],
        [200],
        [300, 400, 500],
    ]
    assert [list(x) for x in array.repartition([4, 5, 5, 8]).partitions] == [
        [0, 1, 2, 100],
        [200],
        [],
        [300, 400, 500],
    ]
    assert [list(x) for x in array.repartition([2, 8]).partitions] == [
        [0, 1],
        [2, 100, 200, 300, 400, 500],
    ]
    assert [list(x) for x in array.repartition([2, 5, 8]).partitions] == [
        [0, 1],
        [2, 100, 200],
        [300, 400, 500],
    ]


def test_getitem_basic():
    one = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.from_iter([[6.6], [], [], [], [7.7, 8.8, 9.9]], highlevel=False)
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    assert ak.to_list(array[0]) == [1.1, 2.2, 3.3]
    assert ak.to_list(array[1]) == []
    assert ak.to_list(array[2]) == [4.4, 5.5]
    assert ak.to_list(array[3]) == [6.6]
    assert ak.to_list(array[4]) == []
    assert ak.to_list(array[5]) == []
    assert ak.to_list(array[6]) == []
    assert ak.to_list(array[7]) == [7.7, 8.8, 9.9]
    assert ak.to_list(array[-1]) == [7.7, 8.8, 9.9]
    assert ak.to_list(array[-2]) == []
    assert ak.to_list(array[-3]) == []
    assert ak.to_list(array[-4]) == []
    assert ak.to_list(array[-5]) == [6.6]
    assert ak.to_list(array[-6]) == [4.4, 5.5]
    assert ak.to_list(array[-7]) == []
    assert ak.to_list(array[-8]) == [1.1, 2.2, 3.3]

    assert (
        array[:].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    )
    assert array[2:6].tojson() == "[[4.4,5.5],[6.6],[],[]]"

    assert array[1:].tojson() == "[[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    assert array[2:].tojson() == "[[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    assert array[3:].tojson() == "[[6.6],[],[],[],[7.7,8.8,9.9]]"
    assert array[4:].tojson() == "[[],[],[],[7.7,8.8,9.9]]"
    assert array[5:].tojson() == "[[],[],[7.7,8.8,9.9]]"
    assert array[6:].tojson() == "[[],[7.7,8.8,9.9]]"
    assert array[7:].tojson() == "[[7.7,8.8,9.9]]"
    assert array[8:].tojson() == "[]"
    assert array[-1:].tojson() == "[[7.7,8.8,9.9]]"
    assert array[-2:].tojson() == "[[],[7.7,8.8,9.9]]"
    assert array[-3:].tojson() == "[[],[],[7.7,8.8,9.9]]"
    assert array[-4:].tojson() == "[[],[],[],[7.7,8.8,9.9]]"
    assert array[-5:].tojson() == "[[6.6],[],[],[],[7.7,8.8,9.9]]"
    assert array[-6:].tojson() == "[[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    assert array[-7:].tojson() == "[[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    assert (
        array[-8:].tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    )

    assert array[:-1].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[]]"
    assert array[:-2].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[]]"
    assert array[:-3].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[]]"
    assert array[:-4].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6]]"
    assert array[:-5].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5]]"
    assert array[:-6].tojson() == "[[1.1,2.2,3.3],[]]"
    assert array[:-7].tojson() == "[[1.1,2.2,3.3]]"
    assert array[:-8].tojson() == "[]"
    assert array[:0].tojson() == "[]"
    assert array[:1].tojson() == "[[1.1,2.2,3.3]]"
    assert array[:2].tojson() == "[[1.1,2.2,3.3],[]]"
    assert array[:3].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5]]"
    assert array[:4].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6]]"
    assert array[:5].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[]]"
    assert array[:6].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[]]"
    assert array[:7].tojson() == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[]]"
    assert (
        array[:8].tojson()
        == "[[1.1,2.2,3.3],[],[4.4,5.5],[6.6],[],[],[],[7.7,8.8,9.9]]"
    )

    one = ak.from_iter(
        [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}],
        highlevel=False,
    )
    two = ak.from_iter(
        [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}], highlevel=False
    )
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    if not ak._util.py27 and not ak._util.py35:
        assert (
            array.tojson()
            == '[{"x":0.0,"y":[]},{"x":1.1,"y":[1]},{"x":2.2,"y":[2,2]},{"x":3.3,"y":[3,3,3]},{"x":4.4,"y":[4,4,4,4]}]'
        )
        assert array["x"].tojson() == "[0.0,1.1,2.2,3.3,4.4]"
        assert array["y"].tojson() == "[[],[1],[2,2],[3,3,3],[4,4,4,4]]"
        assert (
            array[["x"]].tojson()
            == '[{"x":0.0},{"x":1.1},{"x":2.2},{"x":3.3},{"x":4.4}]'
        )


def test_getitem_first_dimension_int():
    one = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.from_iter([[6.6], [], [], [], [7.7, 8.8, 9.9]], highlevel=False)
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    assert ak.to_list(array[0,]) == [1.1, 2.2, 3.3]
    assert ak.to_list(array[-8,]) == [1.1, 2.2, 3.3]
    assert array[0, 1] == 2.2
    assert array[-8, 1] == 2.2
    assert ak.to_list(array[0, [-1, 0]]) == [3.3, 1.1]
    assert ak.to_list(array[0, [False, True, True]]) == [2.2, 3.3]
    assert ak.to_list(array[7,]) == [7.7, 8.8, 9.9]
    assert ak.to_list(array[-1,]) == [7.7, 8.8, 9.9]
    assert array[7, 1] == 8.8
    assert array[-1, 1] == 8.8
    assert ak.to_list(array[7, [-1, 0]]) == [9.9, 7.7]
    assert ak.to_list(array[7, [False, True, True]]) == [8.8, 9.9]


def test_getitem_first_dimension_slice():
    one = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.from_iter([[6.6], [], [], [], [7.7, 8.8, 9.9]], highlevel=False)
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    assert array[2:6,].tojson() == "[[4.4,5.5],[6.6],[],[]]"
    assert (
        array[::-1,].tojson()
        == "[[7.7,8.8,9.9],[],[],[],[6.6],[4.4,5.5],[],[1.1,2.2,3.3]]"
    )
    assert (
        array[::-1, :2].tojson() == "[[7.7,8.8],[],[],[],[6.6],[4.4,5.5],[],[1.1,2.2]]"
    )
    assert array[::-1, :1].tojson() == "[[7.7],[],[],[],[6.6],[4.4],[],[1.1]]"


def test_getitem_first_dimension_field():
    one = ak.from_iter(
        [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}],
        highlevel=False,
    )
    two = ak.from_iter(
        [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}], highlevel=False
    )
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    if not ak._util.py27 and not ak._util.py35:
        assert (
            array.tojson()
            == '[{"x":0.0,"y":[]},{"x":1.1,"y":[1]},{"x":2.2,"y":[2,2]},{"x":3.3,"y":[3,3,3]},{"x":4.4,"y":[4,4,4,4]}]'
        )
        assert array["y", :, :2].tojson() == "[[],[1],[2,2],[3,3],[4,4]]"
        assert (
            array[["y"], :, :2].tojson()
            == '[{"y":[]},{"y":[1]},{"y":[2,2]},{"y":[3,3]},{"y":[4,4]}]'
        )
        assert array[:, "y", :2].tojson() == "[[],[1],[2,2],[3,3],[4,4]]"
        assert array["y", ..., :2].tojson() == "[[],[1],[2,2],[3,3],[4,4]]"
        assert array[np.newaxis, "y", :, :2].tojson() == "[[[],[1],[2,2],[3,3],[4,4]]]"


def test_getitem_first_dimension_intarray():
    one = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.from_iter([[6.6], [], [], [], [7.7, 8.8, 9.9]], highlevel=False)
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    assert isinstance(array[[-1, 3, 2, 0, 2, 3, 7]], ak.layout.ListArray64)
    assert ak.to_list(array[[-1, 3, 2, 0, 2, 3, 7]]) == [
        [7.7, 8.8, 9.9],
        [6.6],
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]

    m1a = ak.from_iter([-1, 3, 2], highlevel=False)
    m2a = ak.from_iter([0, 2, 3, 7], highlevel=False)
    ma = ak.partition.IrregularlyPartitionedArray([m1a, m2a])
    assert isinstance(array[ma], ak.layout.ListArray64)
    assert ak.to_list(array[ma]) == [
        [7.7, 8.8, 9.9],
        [6.6],
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]

    m1b = ak.from_iter([-1, 3, 2, 0, 2], highlevel=False)
    m2b = ak.from_iter([3, 7], highlevel=False)
    mb = ak.partition.IrregularlyPartitionedArray([m1b, m2b])
    assert isinstance(array[mb], ak.layout.ListArray64)
    assert ak.to_list(array[mb]) == [
        [7.7, 8.8, 9.9],
        [6.6],
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]

    assert isinstance(array[[-1, 3, None, 0, 2, 3, 7]], ak.layout.IndexedOptionArray64)
    assert ak.to_list(array[[-1, 3, None, 0, 2, 3, 7]]) == [
        [7.7, 8.8, 9.9],
        [6.6],
        None,
        [1.1, 2.2, 3.3],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]

    m1a = ak.layout.UnmaskedArray(ak.from_iter([-1, 3, 2], highlevel=False))
    m2a = ak.from_iter([0, None, 3, 7], highlevel=False)
    ma = ak.partition.IrregularlyPartitionedArray([m1a, m2a])
    assert isinstance(array[ma], ak.layout.IndexedOptionArray64)
    assert ak.to_list(array[ma]) == [
        [7.7, 8.8, 9.9],
        [6.6],
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [6.6],
        [7.7, 8.8, 9.9],
    ]

    m1b = ak.from_iter([-1, 3, 2, 0, None], highlevel=False)
    m2b = ak.layout.UnmaskedArray(ak.from_iter([3, 7], highlevel=False))
    mb = ak.partition.IrregularlyPartitionedArray([m1b, m2b])
    assert isinstance(array[mb], ak.layout.IndexedOptionArray64)
    assert ak.to_list(array[mb]) == [
        [7.7, 8.8, 9.9],
        [6.6],
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [6.6],
        [7.7, 8.8, 9.9],
    ]


def test_getitem_first_dimension_boolarray():
    one = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.from_iter([[6.6], [], [], [], [7.7, 8.8, 9.9]], highlevel=False)
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    assert isinstance(
        array[[True, False, True, True, False, True, False, True]],
        ak.partition.IrregularlyPartitionedArray,
    )
    assert (
        array[[True, False, True, True, False, True, False, True]].tojson()
        == "[[1.1,2.2,3.3],[4.4,5.5],[6.6],[],[7.7,8.8,9.9]]"
    )

    m1a = ak.from_iter([True, False, True], highlevel=False)
    m2a = ak.from_iter([True, False, True, False, True], highlevel=False)
    ma = ak.partition.IrregularlyPartitionedArray([m1a, m2a])
    assert isinstance(array[ma], ak.partition.IrregularlyPartitionedArray)
    assert array[ma].tojson() == "[[1.1,2.2,3.3],[4.4,5.5],[6.6],[],[7.7,8.8,9.9]]"

    m1b = ak.from_iter([True, False, True, True, False], highlevel=False)
    m2b = ak.from_iter([True, False, True], highlevel=False)
    mb = ak.partition.IrregularlyPartitionedArray([m1b, m2b])
    assert isinstance(array[mb], ak.partition.IrregularlyPartitionedArray)
    assert array[mb].tojson() == "[[1.1,2.2,3.3],[4.4,5.5],[6.6],[],[7.7,8.8,9.9]]"


def test_getitem_first_dimension_jaggedarray():
    one = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.from_iter([[6.6], [], [], [], [7.7, 8.8, 9.9]], highlevel=False)
    array = ak.partition.IrregularlyPartitionedArray([one, two])

    assert isinstance(
        array[[[2, 0], [], [1], [0, 0], [], [], [], [2, 1, 1, 2]]],
        ak.partition.IrregularlyPartitionedArray,
    )
    assert (
        array[[[2, 0], [], [1], [0, 0], [], [], [], [2, 1, 1, 2]]].tojson()
        == "[[3.3,1.1],[],[5.5],[6.6,6.6],[],[],[],[9.9,8.8,8.8,9.9]]"
    )

    m1a = ak.from_iter([[2, 0], [], [1]], highlevel=False)
    m2a = ak.from_iter([[0, 0], [], [], [], [2, 1, 1, 2]], highlevel=False)
    ma = ak.partition.IrregularlyPartitionedArray([m1a, m2a])
    assert isinstance(array[ma], ak.partition.IrregularlyPartitionedArray)
    assert (
        array[ma].tojson()
        == "[[3.3,1.1],[],[5.5],[6.6,6.6],[],[],[],[9.9,8.8,8.8,9.9]]"
    )

    m1b = ak.from_iter([[2, 0], [], [1], [0, 0], []], highlevel=False)
    m2b = ak.from_iter([[], [], [2, 1, 1, 2]], highlevel=False)
    mb = ak.partition.IrregularlyPartitionedArray([m1b, m2b])
    assert isinstance(array[mb], ak.partition.IrregularlyPartitionedArray)
    assert (
        array[mb].tojson()
        == "[[3.3,1.1],[],[5.5],[6.6,6.6],[],[],[],[9.9,8.8,8.8,9.9]]"
    )


def test_highlevel():
    one = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.from_iter([[6.6], [], [], [], [7.7, 8.8, 9.9]], highlevel=False)
    layout = ak.partition.IrregularlyPartitionedArray([one, two])
    array = ak.Array(layout)

    assert ak.to_list(array) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [],
        [],
        [],
        [7.7, 8.8, 9.9],
    ]


def test_mask():
    array = ak.Array([1, 2, 3, 4, 5])
    mask = ak.Array([False, True, True, True, False])
    assert ak.to_list(ak.mask(array, mask)) == [None, 2, 3, 4, None]


def test_indexed():
    content = ak.from_iter(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], highlevel=False
    )
    index = ak.layout.Index64(
        np.array([3, 2, 2, 4, 9, 8, 0, 7, 7, 5, 2, 2, 6, 7, 3], dtype=np.int64)
    )
    indexedarray = ak.layout.IndexedArray64(index, content)

    array = ak.partition.IrregularlyPartitionedArray([indexedarray])
    assert ak.to_list(array) == [
        3.3,
        2.2,
        2.2,
        4.4,
        9.9,
        8.8,
        0.0,
        7.7,
        7.7,
        5.5,
        2.2,
        2.2,
        6.6,
        7.7,
        3.3,
    ]

    array2 = array.repartition([5, 6, 8, 10, 10, 15])
    assert array2.numpartitions == 6
    assert ak.to_list(array2) == [
        3.3,
        2.2,
        2.2,
        4.4,
        9.9,
        8.8,
        0.0,
        7.7,
        7.7,
        5.5,
        2.2,
        2.2,
        6.6,
        7.7,
        3.3,
    ]


def test_repartition():
    array = ak.Array(
        [[], [1.1, 2.2, 3.3], [], [], [4.4, 5.5], [], [6.6], [7.7, 8.8, 9.9], []]
    )
    array2 = ak.repartition(array, 2)
    array3 = ak.repartition(array, 3)
    array4 = ak.repartition(array, [3, 2, 3, 1])
    array5 = ak.repartition(array2, None)

    assert isinstance(array.layout, ak.layout.Content)
    assert isinstance(array2.layout, ak.partition.PartitionedArray)
    assert isinstance(array3.layout, ak.partition.PartitionedArray)
    assert isinstance(array4.layout, ak.partition.PartitionedArray)
    assert isinstance(array5.layout, ak.layout.Content)

    assert ak.to_list(array) == [
        [],
        [1.1, 2.2, 3.3],
        [],
        [],
        [4.4, 5.5],
        [],
        [6.6],
        [7.7, 8.8, 9.9],
        [],
    ]
    assert ak.to_list(array2) == [
        [],
        [1.1, 2.2, 3.3],
        [],
        [],
        [4.4, 5.5],
        [],
        [6.6],
        [7.7, 8.8, 9.9],
        [],
    ]
    assert ak.to_list(array3) == [
        [],
        [1.1, 2.2, 3.3],
        [],
        [],
        [4.4, 5.5],
        [],
        [6.6],
        [7.7, 8.8, 9.9],
        [],
    ]
    assert ak.to_list(array4) == [
        [],
        [1.1, 2.2, 3.3],
        [],
        [],
        [4.4, 5.5],
        [],
        [6.6],
        [7.7, 8.8, 9.9],
        [],
    ]
    assert ak.to_list(array5) == [
        [],
        [1.1, 2.2, 3.3],
        [],
        [],
        [4.4, 5.5],
        [],
        [6.6],
        [7.7, 8.8, 9.9],
        [],
    ]

    assert [ak.to_list(x) for x in array2.layout.partitions] == [
        [[], [1.1, 2.2, 3.3]],
        [[], []],
        [[4.4, 5.5], []],
        [[6.6], [7.7, 8.8, 9.9]],
        [[]],
    ]
    assert [ak.to_list(x) for x in array3.layout.partitions] == [
        [[], [1.1, 2.2, 3.3], []],
        [[], [4.4, 5.5], []],
        [[6.6], [7.7, 8.8, 9.9], []],
    ]
    assert [ak.to_list(x) for x in array4.layout.partitions] == [
        [[], [1.1, 2.2, 3.3], []],
        [[], [4.4, 5.5]],
        [[], [6.6], [7.7, 8.8, 9.9]],
        [[]],
    ]


def test_firsts_singletons():
    array = ak.Array([None, 1.1, 2.2, None, 3.3, None, None, 4.4, 5.5, None])

    one = ak.singletons(array)
    assert ak.to_list(one) == [[], [1.1], [2.2], [], [3.3], [], [], [4.4], [5.5], []]
    two = ak.firsts(one)
    assert ak.to_list(two) == [None, 1.1, 2.2, None, 3.3, None, None, 4.4, 5.5, None]

    array = ak.repartition(array, 3)
    assert isinstance(array.layout, ak.partition.PartitionedArray)

    one = ak.singletons(array)
    assert isinstance(one.layout, ak.partition.PartitionedArray)
    assert ak.to_list(one) == [[], [1.1], [2.2], [], [3.3], [], [], [4.4], [5.5], []]
    two = ak.firsts(one)
    assert isinstance(two.layout, ak.partition.PartitionedArray)
    assert ak.to_list(two) == [None, 1.1, 2.2, None, 3.3, None, None, 4.4, 5.5, None]


def test_mask2():
    array = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    assert isinstance(array.layout, ak.layout.Content)
    mask = ak.Array([False, False, True, True, False, True, True, False, True])
    assert isinstance(mask.layout, ak.layout.Content)

    one = array.mask[mask]
    assert isinstance(one.layout, ak.layout.Content)
    assert ak.to_list(one) == [None, None, 3.3, 4.4, None, 6.6, 7.7, None, 9.9]

    array = ak.repartition(array, 4)
    assert isinstance(array.layout, ak.partition.PartitionedArray)

    one = array.mask[mask]
    assert isinstance(one.layout, ak.partition.PartitionedArray)
    assert ak.to_list(one) == [None, None, 3.3, 4.4, None, 6.6, 7.7, None, 9.9]

    mask = ak.repartition(mask, 3)
    assert isinstance(mask.layout, ak.partition.PartitionedArray)

    one = array.mask[mask]
    assert isinstance(one.layout, ak.partition.PartitionedArray)
    assert ak.to_list(one) == [None, None, 3.3, 4.4, None, 6.6, 7.7, None, 9.9]

    array = ak.repartition(array, None)
    assert isinstance(array.layout, ak.layout.Content)

    one = array.mask[mask]
    assert isinstance(one.layout, ak.partition.PartitionedArray)
    assert ak.to_list(one) == [None, None, 3.3, 4.4, None, 6.6, 7.7, None, 9.9]

    mask = ak.repartition(mask, None)
    assert isinstance(mask.layout, ak.layout.Content)

    one = array.mask[mask]
    assert isinstance(one.layout, ak.layout.Content)
    assert ak.to_list(one) == [None, None, 3.3, 4.4, None, 6.6, 7.7, None, 9.9]


def test_zip():
    x = ak.Array([[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]])
    y = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5])

    one = ak.zip({"x": x, "y": y})
    two = ak.zip({"x": x, "y": y}, depth_limit=1)
    xx, yy = ak.unzip(two)
    assert isinstance(one.layout, ak.layout.Content)
    assert isinstance(two.layout, ak.layout.Content)
    assert isinstance(xx.layout, ak.layout.Content)
    assert isinstance(yy.layout, ak.layout.Content)
    assert ak.to_list(one) == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": 1.1}, {"x": 3, "y": 1.1}],
        [],
        [{"x": 4, "y": 3.3}, {"x": 5, "y": 3.3}],
        [{"x": 6, "y": 4.4}],
        [
            {"x": 7, "y": 5.5},
            {"x": 8, "y": 5.5},
            {"x": 9, "y": 5.5},
            {"x": 10, "y": 5.5},
        ],
    ]
    assert ak.to_list(two) == [
        {"x": [1, 2, 3], "y": 1.1},
        {"x": [], "y": 2.2},
        {"x": [4, 5], "y": 3.3},
        {"x": [6], "y": 4.4},
        {"x": [7, 8, 9, 10], "y": 5.5},
    ]
    if not ak._util.py27 and not ak._util.py35:
        assert ak.to_list(xx) == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]
        assert ak.to_list(yy) == [1.1, 2.2, 3.3, 4.4, 5.5]

    x = ak.repartition(x, 3)
    assert isinstance(x.layout, ak.partition.PartitionedArray)
    assert ak.to_list(x) == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]

    one = ak.zip({"x": x, "y": y})
    two = ak.zip({"x": x, "y": y}, depth_limit=1)
    xx, yy = ak.unzip(two)
    assert isinstance(one.layout, ak.partition.PartitionedArray)
    assert isinstance(two.layout, ak.partition.PartitionedArray)
    assert isinstance(xx.layout, ak.partition.PartitionedArray)
    assert isinstance(yy.layout, ak.partition.PartitionedArray)
    assert ak.to_list(one) == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": 1.1}, {"x": 3, "y": 1.1}],
        [],
        [{"x": 4, "y": 3.3}, {"x": 5, "y": 3.3}],
        [{"x": 6, "y": 4.4}],
        [
            {"x": 7, "y": 5.5},
            {"x": 8, "y": 5.5},
            {"x": 9, "y": 5.5},
            {"x": 10, "y": 5.5},
        ],
    ]
    assert ak.to_list(two) == [
        {"x": [1, 2, 3], "y": 1.1},
        {"x": [], "y": 2.2},
        {"x": [4, 5], "y": 3.3},
        {"x": [6], "y": 4.4},
        {"x": [7, 8, 9, 10], "y": 5.5},
    ]
    if not ak._util.py27 and not ak._util.py35:
        assert ak.to_list(xx) == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]
        assert ak.to_list(yy) == [1.1, 2.2, 3.3, 4.4, 5.5]

    y = ak.repartition(y, 2)
    assert isinstance(x.layout, ak.partition.PartitionedArray)
    assert ak.to_list(y) == [1.1, 2.2, 3.3, 4.4, 5.5]

    one = ak.zip({"x": x, "y": y})
    two = ak.zip({"x": x, "y": y}, depth_limit=1)
    xx, yy = ak.unzip(two)
    assert isinstance(one.layout, ak.partition.PartitionedArray)
    assert isinstance(two.layout, ak.partition.PartitionedArray)
    assert isinstance(xx.layout, ak.partition.PartitionedArray)
    assert isinstance(yy.layout, ak.partition.PartitionedArray)
    assert ak.to_list(one) == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": 1.1}, {"x": 3, "y": 1.1}],
        [],
        [{"x": 4, "y": 3.3}, {"x": 5, "y": 3.3}],
        [{"x": 6, "y": 4.4}],
        [
            {"x": 7, "y": 5.5},
            {"x": 8, "y": 5.5},
            {"x": 9, "y": 5.5},
            {"x": 10, "y": 5.5},
        ],
    ]
    assert ak.to_list(two) == [
        {"x": [1, 2, 3], "y": 1.1},
        {"x": [], "y": 2.2},
        {"x": [4, 5], "y": 3.3},
        {"x": [6], "y": 4.4},
        {"x": [7, 8, 9, 10], "y": 5.5},
    ]
    if not ak._util.py27 and not ak._util.py35:
        assert ak.to_list(xx) == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]
        assert ak.to_list(yy) == [1.1, 2.2, 3.3, 4.4, 5.5]

    x = ak.repartition(x, None)
    assert isinstance(x.layout, ak.layout.Content)
    assert ak.to_list(x) == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]

    one = ak.zip({"x": x, "y": y})
    two = ak.zip({"x": x, "y": y}, depth_limit=1)
    xx, yy = ak.unzip(two)
    assert isinstance(one.layout, ak.partition.PartitionedArray)
    assert isinstance(two.layout, ak.partition.PartitionedArray)
    assert isinstance(xx.layout, ak.partition.PartitionedArray)
    assert isinstance(yy.layout, ak.partition.PartitionedArray)
    assert ak.to_list(one) == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": 1.1}, {"x": 3, "y": 1.1}],
        [],
        [{"x": 4, "y": 3.3}, {"x": 5, "y": 3.3}],
        [{"x": 6, "y": 4.4}],
        [
            {"x": 7, "y": 5.5},
            {"x": 8, "y": 5.5},
            {"x": 9, "y": 5.5},
            {"x": 10, "y": 5.5},
        ],
    ]
    assert ak.to_list(two) == [
        {"x": [1, 2, 3], "y": 1.1},
        {"x": [], "y": 2.2},
        {"x": [4, 5], "y": 3.3},
        {"x": [6], "y": 4.4},
        {"x": [7, 8, 9, 10], "y": 5.5},
    ]
    if not ak._util.py27 and not ak._util.py35:
        assert ak.to_list(xx) == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]
        assert ak.to_list(yy) == [1.1, 2.2, 3.3, 4.4, 5.5]

    y = ak.repartition(y, None)
    assert isinstance(y.layout, ak.layout.Content)
    assert ak.to_list(y) == [1.1, 2.2, 3.3, 4.4, 5.5]

    one = ak.zip({"x": x, "y": y})
    two = ak.zip({"x": x, "y": y}, depth_limit=1)
    xx, yy = ak.unzip(two)
    assert isinstance(one.layout, ak.layout.Content)
    assert isinstance(two.layout, ak.layout.Content)
    assert isinstance(xx.layout, ak.layout.Content)
    assert isinstance(yy.layout, ak.layout.Content)
    assert ak.to_list(one) == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": 1.1}, {"x": 3, "y": 1.1}],
        [],
        [{"x": 4, "y": 3.3}, {"x": 5, "y": 3.3}],
        [{"x": 6, "y": 4.4}],
        [
            {"x": 7, "y": 5.5},
            {"x": 8, "y": 5.5},
            {"x": 9, "y": 5.5},
            {"x": 10, "y": 5.5},
        ],
    ]
    assert ak.to_list(two) == [
        {"x": [1, 2, 3], "y": 1.1},
        {"x": [], "y": 2.2},
        {"x": [4, 5], "y": 3.3},
        {"x": [6], "y": 4.4},
        {"x": [7, 8, 9, 10], "y": 5.5},
    ]
    if not ak._util.py27 and not ak._util.py35:
        assert ak.to_list(xx) == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]
        assert ak.to_list(yy) == [1.1, 2.2, 3.3, 4.4, 5.5]


def test_with_name_field():
    array = ak.Array(
        [
            {"x": 0.0, "y": []},
            {"x": 1.1, "y": [1]},
            {"x": 2.2, "y": [2, 2]},
            {"x": 3.3, "y": [3, 3, 3]},
        ]
    )
    array2 = ak.repartition(array, 2)
    z = ak.Array([100, 200, 300, 400])
    z2 = ak.repartition(z, 3)

    one = ak.with_name(array, "Wilbur")
    assert isinstance(one.layout, ak.layout.Content)
    assert one.layout.parameters["__record__"] == "Wilbur"

    one = ak.with_name(array2, "Wilbur")
    assert isinstance(one.layout, ak.partition.PartitionedArray)
    assert one.layout.partition(0).parameters["__record__"] == "Wilbur"
    assert one.layout.partition(1).parameters["__record__"] == "Wilbur"

    two = ak.with_field(array, z, "z")
    assert isinstance(two.layout, ak.layout.Content)
    assert ak.to_list(two) == [
        {"x": 0.0, "y": [], "z": 100},
        {"x": 1.1, "y": [1], "z": 200},
        {"x": 2.2, "y": [2, 2], "z": 300},
        {"x": 3.3, "y": [3, 3, 3], "z": 400},
    ]

    two = ak.with_field(array2, z, "z")
    assert isinstance(two.layout, ak.partition.PartitionedArray)
    assert ak.to_list(two) == [
        {"x": 0.0, "y": [], "z": 100},
        {"x": 1.1, "y": [1], "z": 200},
        {"x": 2.2, "y": [2, 2], "z": 300},
        {"x": 3.3, "y": [3, 3, 3], "z": 400},
    ]

    two = ak.with_field(array2, z2, "z")
    assert isinstance(two.layout, ak.partition.PartitionedArray)
    assert ak.to_list(two) == [
        {"x": 0.0, "y": [], "z": 100},
        {"x": 1.1, "y": [1], "z": 200},
        {"x": 2.2, "y": [2, 2], "z": 300},
        {"x": 3.3, "y": [3, 3, 3], "z": 400},
    ]

    two = ak.with_field(array2, z2, "z")
    assert isinstance(two.layout, ak.partition.PartitionedArray)
    assert ak.to_list(two) == [
        {"x": 0.0, "y": [], "z": 100},
        {"x": 1.1, "y": [1], "z": 200},
        {"x": 2.2, "y": [2, 2], "z": 300},
        {"x": 3.3, "y": [3, 3, 3], "z": 400},
    ]


def test_atleast_1d():
    array = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5])
    array2 = ak.repartition(array, 2)

    one = ak.atleast_1d(array)
    assert isinstance(one, np.ndarray)
    assert ak.to_list(one) == [1.1, 2.2, 3.3, 4.4, 5.5]

    one = ak.atleast_1d(array2)
    assert isinstance(one, np.ndarray)
    assert ak.to_list(one) == [1.1, 2.2, 3.3, 4.4, 5.5]


def test_0167_strings():
    array = ak.repartition(
        ak.Array(["one", "two", "three", "two", "two", "one", "three"]), 3
    )

    assert ak.to_list(array == "two") == [False, True, False, True, True, False, False]
    assert ak.to_list("two" == array) == [False, True, False, True, True, False, False]
    assert ak.to_list(array == ["two"]) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]
    assert ak.to_list(["two"] == array) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]
    assert ak.to_list(array == ak.Array(["two"])) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]
    assert ak.to_list(ak.Array(["two"]) == array) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]

    array = ak.Array([["one", "two", "three"], [], ["two"], ["two", "one"], ["three"]])
    assert ak.to_list(array == "two") == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list("two" == array) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(array == ["two"]) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(["two"] == array) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(array == ak.Array(["two"])) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(ak.Array(["two"]) == array) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]

    array = ak.Array([["one", "two", "three"], [], ["two"], ["two", "one"], ["three"]])
    assert ak.to_list(array == ["three", "two", "one", "one", "three"]) == [
        [False, False, True],
        [],
        [False],
        [False, True],
        [True],
    ]
    assert ak.to_list(["three", "two", "one", "one", "three"] == array) == [
        [False, False, True],
        [],
        [False],
        [False, True],
        [True],
    ]
    assert ak.to_list(array == ak.Array(["three", "two", "one", "one", "three"])) == [
        [False, False, True],
        [],
        [False],
        [False, True],
        [True],
    ]
    assert ak.to_list(ak.Array(["three", "two", "one", "one", "three"]) == array) == [
        [False, False, True],
        [],
        [False],
        [False, True],
        [True],
    ]
